package net.lamgc.scext.onedrive_transfer.remote

import com.microsoft.graph.models.DriveItemCreateUploadSessionParameterSet
import com.microsoft.graph.models.DriveItemUploadableProperties
import io.ktor.http.*
import io.ktor.server.application.*
import io.ktor.server.auth.*
import io.ktor.server.engine.*
import io.ktor.server.netty.*
import io.ktor.server.response.*
import io.ktor.server.routing.*
import io.ktor.util.collections.*
import mu.KotlinLogging
import net.lamgc.scext.onedrive_transfer.*
import net.lamgc.scext.onedrive_transfer.remote.request.*
import okhttp3.internal.toImmutableMap
import org.jetbrains.exposed.sql.SqlExpressionBuilder.greaterEq
import java.time.LocalDateTime
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.ConcurrentLinkedDeque

class RemoteOneDriveTransferExecutorController(
    private val config: CentralSetting,
    private val callback: OneDriveTransferCallback
) : OneDriveTransferTaskExecutor {

    private val logger = KotlinLogging.logger { }

    private val pendingTasks = ConcurrentLinkedDeque<OneDriveTransferTask>()
    private val workingTasks = ConcurrentHashMap<Int, OneDriveTransferWorkerProgress>()
    private val preAllocatedTasks = ConcurrentHashMap<OneDriveTransferTask, PreAllocateTaskEntry>()
    private val cancelledTasks = ConcurrentSet<OneDriveTransferTask>()

    init {
        embeddedServer(Netty, port = config.port) {
            install(Authentication) {
                bearer(AUTHENTICATE_CONFIG_ID) {
                    authenticate {
                        val worker =
                            RegisteredRemoteTransferWorker.find { RegisteredRemoteTransferWorkers.token eq it.token }
                                .limit(1)
                                .firstOrNull()

                        if (worker != null) {
                            worker.lastContactAt = LocalDateTime.now()
                        }
                        return@authenticate worker
                    }
                }
            }

            registerRemoteWorkerRegisterRouting()
            registerTaskAllocationRouting()
            registerTaskProgressUpdateRouting()
            registerFileUploadSessionRequestRouting()
        }.start(wait = false)
    }

    override fun submitTransferTask(task: OneDriveTransferTask): Boolean {
        pendingTasks.addLast(task)
        try {
            callback.onTransferTaskCreated(task)
        } catch (e: Exception) {
            logger.error(e) { "Error occurred while calling onTransferTaskCreated." }
        }
        return true
    }

    override fun cancelTransferTask(task: OneDriveTransferTask): Boolean {
        if (pendingTasks.remove(task)) {
            try {
                callback.onTransferCancelled(task, null)
            } catch (e: Exception) {
                logger.error(e) { "Error occurred while calling onTransferCancelled." }
            }
            return true
        }

        if (workingTasks.values.any { it.currentTask == task }) {
            cancelledTasks.add(task)
            return true
        } else if (preAllocatedTasks.contains(task)) {
            preAllocatedTasks.remove(task)
            try {
                callback.onTransferCancelled(task, null)
            } catch (e: Exception) {
                logger.error(e) { "Error occurred while calling onTransferCancelled." }
            }
            return true
        }

        return false
    }

    override fun getQueuedTransferTasks(): List<OneDriveTransferTask> {
        return this.pendingTasks.toList()
    }

    override fun getWorkerCount(): Int {
        // 在近 1 小时内有联系的 Worker 数量
        val minusHours = LocalDateTime.now().minusHours(1)
        return RegisteredRemoteTransferWorker
            .count(RegisteredRemoteTransferWorkers.lastContactAt greaterEq minusHours).toInt()
    }

    override fun getWorkingTasks(): Map<Int, OneDriveTransferWorkerProgress> {
        return workingTasks.toImmutableMap()
    }

    private fun Application.registerRemoteWorkerRegisterRouting() {
        routing {
            post<RegisterWorkerRequest> {
                if (config.registerToken != it.token) {
                    call.respond(HttpStatusCode.Forbidden, "Invalid token.")
                    return@post
                }

                val existsWorker =
                    RegisteredRemoteTransferWorker.find { RegisteredRemoteTransferWorkers.workerName eq it.workerName }
                        .firstOrNull()

                val workerToken = randomString(128)
                if (existsWorker != null) {
                    val now = LocalDateTime.now()
                    // 如果 existsWorker 的 lastContactAt 距离现在不超过 12 小时, 则拒绝重新注册
                    if (existsWorker.lastContactAt > now.minusHours(12)) {
                        call.respond(HttpStatusCode.Forbidden, "Worker already registered.")
                        return@post
                    } else {
                        existsWorker.token = workerToken
                        existsWorker.registeredAt = now
                        existsWorker.lastContactAt = now
                    }
                } else {
                    RegisteredRemoteTransferWorker.new {
                        this.workerName = it.workerName
                        this.registeredAt = LocalDateTime.now()
                        this.lastContactAt = LocalDateTime.now()
                        this.token = workerToken
                    }
                }

                call.respond(HttpStatusCode.OK, RegisterWorkerResponse(true, "Worker registered.", workerToken))
            }
        }
    }

    private fun Application.registerTaskAllocationRouting() {
        routing {
            authenticate(AUTHENTICATE_CONFIG_ID) {
                post<TaskAllocationRequest> { request ->
                    val worker = call.authentication.principal<RegisteredRemoteTransferWorker>()
                        ?: return@post call.respond(HttpStatusCode.Unauthorized)

                    // 寻找 pendingTasks 中 fileSize 不超过 worker 的限制的任务
                    val task = pendingTasks.find { task ->
                        task.document.fileSize <= request.acceptMaxSize && task !in cancelledTasks && preAllocatedTasks.computeIfAbsent(
                            task
                        ) {
                            PreAllocateTaskEntry(task, worker)
                        }.worker == worker
                    } ?: return@post call.respond(HttpStatusCode.NoContent)

                    call.respond(
                        HttpStatusCode.OK, TaskAllocationResponse(
                            true, TaskAllocationInfo(
                                taskId = task.id,
                                ownerBotToken = "",
                                fileId = task.document.fileId,
                                fileSize = task.document.fileSize
                            )
                        )
                    )
                }

                post<TaskAcceptRequest> {
                    val worker = call.authentication.principal<RegisteredRemoteTransferWorker>()
                        ?: return@post call.respond(HttpStatusCode.Unauthorized)

                    val preAllocatedTaskEntry = preAllocatedTasks.entries.find { entry ->
                        entry.key.id == it.taskId && entry.value.worker.id == worker.id
                    } ?: return@post call.respond(HttpStatusCode.OK, TaskAcceptResponse(false))

                    pendingTasks.remove(preAllocatedTaskEntry.key)
                    val existsTask = workingTasks[worker.id.value]
                    if (existsTask != null) {
                        if (cancelledTasks.remove(existsTask.currentTask)) {
                            try {
                                callback.onTransferCancelled(existsTask.currentTask, existsTask)
                            } catch (e: Exception) {
                                logger.error(e) { "Error occurred while calling onTransferCancelled." }
                            }
                        } else {
                            pendingTasks.addFirst(existsTask.currentTask)
                        }
                    }
                    val workerProgress = OneDriveTransferWorkerProgress(preAllocatedTaskEntry.key)

                    workingTasks[worker.id.value] = workerProgress
                    preAllocatedTasks.remove(preAllocatedTaskEntry.key)

                    try {
                        callback.onTransferTaskStart(workerProgress)
                    } catch (e: Exception) {
                        logger.error(e) { "Error occurred while calling onTransferTaskStart." }
                    }

                    call.respond(HttpStatusCode.OK, TaskAcceptResponse(true))
                }
            }
        }
    }

    private fun Application.registerTaskProgressUpdateRouting() {
        routing {
            authenticate(AUTHENTICATE_CONFIG_ID) {
                post<TaskProgressUpdateRequest> {
                    val worker = call.authentication.principal<RegisteredRemoteTransferWorker>()
                        ?: return@post call.respond(HttpStatusCode.Unauthorized)

                    val processingTask = workingTasks[worker.id.value]
                    if (processingTask == null || processingTask.currentTask.id != it.taskId) {
                        return@post call.respond(HttpStatusCode.BadRequest)
                    }

                    processingTask.progress.set(it.progress)
                    processingTask.status = when (it.status) {
                        TaskStatus.DOWNLOADING_FILE -> OneDriveTransferStatus.DOWNLOADING_FILE
                        TaskStatus.DOWNLOADED_FILE -> OneDriveTransferStatus.CREATING_UPLOAD_SESSION
                        TaskStatus.UPLOADING_FILE -> OneDriveTransferStatus.UPLOADING
                        TaskStatus.COMPLETED -> OneDriveTransferStatus.SUCCESS
                    }

                    if (cancelledTasks.contains(processingTask.currentTask)) {
                        cancelledTasks.remove(processingTask.currentTask)
                        workingTasks.remove(worker.id.value)
                        try {
                            callback.onTransferCancelled(processingTask.currentTask, processingTask)
                        } catch (e: Exception) {
                            logger.error(e) { "Error occurred while calling onTransferCancelled." }
                        }
                        return@post call.respond(
                            HttpStatusCode.OK,
                            TaskProgressUpdateResponse(TaskProgressUpdateStatus.CANCELLED)
                        )
                    }

                    try {
                        callback.onUploadProgress(processingTask)
                    } catch (e: Exception) {
                        logger.error(e) { "Error occurred while calling onUploadProgress." }
                    }

                    if (processingTask.status == OneDriveTransferStatus.SUCCESS) {
                        try {
                            callback.onTransferSuccess(processingTask.currentTask, processingTask)
                        } catch (e: Exception) {
                            logger.error(e) { "Error occurred while calling onTransferSuccess." }
                        }
                    }

                    return@post call.respond(
                        HttpStatusCode.OK,
                        TaskProgressUpdateResponse(TaskProgressUpdateStatus.CONTINUE)
                    )
                }
            }
        }
    }

    private fun Application.registerFileUploadSessionRequestRouting() {
        routing {
            authenticate(AUTHENTICATE_CONFIG_ID) {
                post<FileUploadSessionRequest> {
                    val worker = call.authentication.principal<RegisteredRemoteTransferWorker>()
                        ?: return@post call.respond(HttpStatusCode.Unauthorized)

                    val processingTask = workingTasks[worker.id.value]
                    if (processingTask == null || processingTask.currentTask.id != it.taskId) {
                        return@post call.respond(HttpStatusCode.BadRequest)
                    }

                    if (processingTask.status != OneDriveTransferStatus.CREATING_UPLOAD_SESSION) {
                        return@post call.respond(HttpStatusCode.BadRequest)
                    }

                    val task = processingTask.currentTask
                    val file = task.document

                    val graphClient = task.service.createGraphClient(task.tgUserId)
                    val drive = graphClient.drives(task.onedriveId).buildRequest().get()
                        ?: throw IllegalStateException("无法获取 OneDrive 驱动器.")
                    if (file.fileSize > drive.quota!!.remaining!!) {
                        throw IllegalStateException("OneDrive 剩余空间不足.")
                    }

                    val filePath = OneDriveUtils.checkAndGetPath(
                        graphClient,
                        task.onedriveId,
                        task.storagePath,
                        task.document.fileName
                    )
                    logger.debug { "OneDrive 中转任务: ${task.document.fileName} -> $filePath" }

                    val uploadSession = graphClient.drives(task.onedriveId).root().itemWithPath(filePath)
                        .createUploadSession(
                            DriveItemCreateUploadSessionParameterSet.newBuilder()
                                .withItem(DriveItemUploadableProperties().apply {
                                    fileSize = file.fileSize
                                    name = task.document.fileName
                                })
                                .build()
                        )
                        .buildRequest()
                        .post() ?: throw IllegalStateException("无法创建 OneDrive 上传会话.")

                    val uploadUrl =
                        uploadSession.uploadUrl ?: return@post call.respond(HttpStatusCode.InternalServerError)

                    return@post call.respond(HttpStatusCode.OK, FileUploadSessionResponse(uploadUrl))
                }
            }
        }
    }

    data class PreAllocateTaskEntry(
        val task: OneDriveTransferTask,
        val worker: RegisteredRemoteTransferWorker,
        val allocatedAt: LocalDateTime = LocalDateTime.now()
    )

    companion object {
        private const val AUTHENTICATE_CONFIG_ID = "worker-token"
    }

}
